@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: main.c 8673 2009-06-11 22:36:45Z charles $
+ * $Id: main.c 9358 2009-10-22 02:25:26Z charles $
  *
  * Copyright (c) 2005-2008 Transmission authors and contributors
  *
@@ -88,10 +88,6 @@ struct cbdata
     GtkTreeSelection  * sel;
 };
 
-#define CBDATA_PTR "callback-data-pointer"
-
-static GtkUIManager * myUIManager = NULL;
-
 static void           appsetup( TrWindow * wind,
                                 GSList *   args,
                                 struct     cbdata *,
@@ -101,7 +97,7 @@ static void           appsetup( TrWindow * wind,
 static void           winsetup( struct cbdata * cbdata,
                                 TrWindow *      wind );
 
-static void           wannaquit( void * vdata );
+static void           wannaquit( gpointer vdata );
 
 static void           setupdrag( GtkWidget *    widget,
                                  struct cbdata *data );
@@ -180,6 +176,7 @@ refreshActions( struct cbdata * data )
     action_sensitize( "remove-torrent", counts.totalCount != 0 );
     action_sensitize( "delete-torrent", counts.totalCount != 0 );
     action_sensitize( "verify-torrent", counts.totalCount != 0 );
+    action_sensitize( "show-torrent-properties", counts.totalCount != 0 );
     action_sensitize( "open-torrent-folder", counts.totalCount == 1 );
     action_sensitize( "relocate-torrent", counts.totalCount == 1 );
 
@@ -262,32 +259,29 @@ onMainWindowSizeAllocated( GtkWidget *            window,
 }
 
 static sig_atomic_t global_sigcount = 0;
+static struct cbdata * sighandler_cbdata = NULL;
 
 static void
-fatalsig( int sig )
+signal_handler( int sig )
 {
-    /* revert to default handler after this many */
-    static const int SIGCOUNT_MAX = 3;
-
-    if( ++global_sigcount >= SIGCOUNT_MAX )
+    if( ++global_sigcount > 1 )
     {
         signal( sig, SIG_DFL );
         raise( sig );
     }
+    else switch( sig )
+    {
+        case SIGINT:
+            g_message( _( "Got signal %d; trying to shut down cleanly.  Do it again if it gets stuck." ), sig );
+            doAction( "quit", sighandler_cbdata );
+            break;
+    }
 }
 
 static void
 setupsighandlers( void )
 {
-#ifdef G_OS_WIN32
-    const int sigs[] = { SIGINT, SIGTERM };
-#else
-    const int sigs[] = { SIGHUP, SIGINT, SIGQUIT, SIGTERM };
-#endif
-    guint     i;
-
-    for( i = 0; i < G_N_ELEMENTS( sigs ); ++i )
-        signal( sigs[i], fatalsig );
+    signal( SIGINT, signal_handler );
 }
 
 static tr_rpc_callback_status
@@ -328,20 +322,19 @@ onRPCChanged( tr_session            * session UNUSED,
 }
 
 int
-main( int     argc,
-      char ** argv )
+main( int argc, char ** argv )
 {
-    char *              err = NULL;
-    GSList *            argfiles;
-    GError *            gerr;
-    gboolean            didinit = FALSE;
-    gboolean            didlock = FALSE;
-    gboolean            showversion = FALSE;
-    gboolean            startpaused = FALSE;
-    gboolean            startminimized = FALSE;
-    const char *        domain = MY_NAME;
-    char *              configDir = NULL;
-    tr_lockfile_state_t tr_state;
+    char * err = NULL;
+    GSList * argfiles;
+    GError * gerr;
+    gboolean didinit = FALSE;
+    gboolean didlock = FALSE;
+    gboolean showversion = FALSE;
+    gboolean startpaused = FALSE;
+    gboolean startminimized = FALSE;
+    const char * domain = MY_NAME;
+    char * configDir = NULL;
+    gtr_lockfile_state_t tr_state;
 
     GOptionEntry entries[] = {
         { "paused",     'p', 0, G_OPTION_ARG_NONE,
@@ -370,7 +363,7 @@ main( int     argc,
         g_thread_init( NULL );
 
     gerr = NULL;
-    if( !gtk_init_with_args( &argc, &argv, _( "[torrent files]" ), entries,
+    if( !gtk_init_with_args( &argc, &argv, (char*)_( "[torrent files]" ), entries,
                              (char*)domain, &gerr ) )
     {
         fprintf( stderr, "%s\n", gerr->message );
@@ -418,12 +411,13 @@ main( int     argc,
             }
         }
     }
-    else if( ( !didlock ) && ( tr_state == TR_LOCKFILE_ELOCK ) )
+    else if( ( !didlock ) && ( tr_state == GTR_LOCKFILE_ELOCK ) )
     {
         /* There's already another copy of Transmission running,
          * so tell it to present its window to the user */
-        gtr_dbus_present_window( );
         err = NULL;
+        if( !gtr_dbus_present_window( ) )
+            err = g_strdup( _( "Transmission is already running, but is not responding.  To start a new session, you must first close the existing Transmission process." ) );
     }
 
     if( didlock && ( didinit || cf_init( configDir, &err ) ) )
@@ -433,14 +427,17 @@ main( int     argc,
 
         const char * str;
         GtkWindow * win;
+        GtkUIManager * myUIManager;
         tr_session * session;
         struct cbdata * cbdata = g_new0( struct cbdata, 1 );
 
+        sighandler_cbdata = cbdata;
+
         /* ensure the directories are created */
-       if(( str = pref_string_get( PREF_KEY_DIR_WATCH )))
-           mkdir_p( str, 0777 );
-       if(( str = pref_string_get( TR_PREFS_KEY_DOWNLOAD_DIR )))
-           mkdir_p( str, 0777 );
+        if(( str = pref_string_get( PREF_KEY_DIR_WATCH )))
+            gtr_mkdir_with_parents( str, 0777 );
+        if(( str = pref_string_get( TR_PREFS_KEY_DOWNLOAD_DIR )))
+            gtr_mkdir_with_parents( str, 0777 );
 
         /* initialize the libtransmission session */
         session = tr_sessionInit( "gtk", configDir, TRUE, pref_get_all( ) );
@@ -471,8 +468,11 @@ main( int     argc,
     }
     else if( err )
     {
-        gtk_widget_show( errmsg_full( NULL, (callbackfunc_t)gtk_main_quit,
-                                      NULL, "%s", err ) );
+        const char * primary_text = _( "Transmission cannot be started." );
+        GtkWidget * w = gtk_message_dialog_new( NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, primary_text, NULL );
+        gtk_message_dialog_format_secondary_text( GTK_MESSAGE_DIALOG( w ), "%s", err );
+        g_signal_connect( w, "response", G_CALLBACK(gtk_main_quit), NULL );
+        gtk_widget_show( w );
         g_free( err );
         gtk_main( );
     }
@@ -537,9 +537,10 @@ appsetup( TrWindow *      wind,
         gtk_widget_show( GTK_WIDGET( wind ) );
     else
     {
-        gtk_window_iconify( wind );
         gtk_window_set_skip_taskbar_hint( cbdata->wind,
                                           cbdata->icon != NULL );
+        cbdata->isIconified = FALSE; // ensure that the next toggle iconifies
+        action_toggle( "toggle-main-window", FALSE );
     }
 }
 
@@ -630,6 +631,7 @@ static gpointer
 quitThreadFunc( gpointer gdata )
 {
     struct cbdata * cbdata = gdata;
+    gdk_threads_enter( );
 
     tr_core_close( cbdata->core );
 
@@ -656,6 +658,8 @@ quitThreadFunc( gpointer gdata )
     g_free( cbdata );
 
     gtk_main_quit( );
+    gdk_threads_leave( );
+
     return NULL;
 }
 
@@ -667,7 +671,7 @@ do_exit_cb( GtkWidget *w  UNUSED,
 }
 
 static void
-wannaquit( void * vdata )
+wannaquit( gpointer vdata )
 {
     GtkWidget *r, *p, *b, *w, *c;
     struct cbdata *cbdata = vdata;
@@ -828,7 +832,7 @@ setupdrag( GtkWidget *    widget,
                           gotdrag ), data );
 
     gtk_drag_dest_set( widget, GTK_DEST_DEFAULT_ALL, targets,
-                       ALEN( targets ), GDK_ACTION_COPY | GDK_ACTION_MOVE );
+                       G_N_ELEMENTS( targets ), GDK_ACTION_COPY | GDK_ACTION_MOVE );
 }
 
 static void
@@ -887,12 +891,12 @@ coreerr( TrCore * core UNUSED, guint code, const char * msg, struct cbdata * c )
 {
     switch( code )
     {
-        case TR_EINVALID:
+        case TR_PARSE_ERR:
             c->errqueue =
                 g_slist_append( c->errqueue, g_path_get_basename( msg ) );
             break;
 
-        case TR_EDUPLICATE:
+        case TR_PARSE_DUPLICATE:
             c->dupqueue = g_slist_append( c->dupqueue, g_strdup( msg ) );
             break;
 
@@ -900,10 +904,6 @@ coreerr( TrCore * core UNUSED, guint code, const char * msg, struct cbdata * c )
             showTorrentErrors( c );
             break;
 
-        case TR_CORE_ERR_SAVE_STATE:
-            errmsg( c->wind, "%s", msg );
-            break;
-
         default:
             g_assert_not_reached( );
             break;
@@ -1329,6 +1329,10 @@ doAction( const char * action_name, gpointer user_data )
         GtkWidget * dialog = stats_dialog_create( data->wind, data->core );
         gtk_widget_show( dialog );
     }
+    else if( !strcmp( action_name, "donate" ) )
+    {
+        gtr_open_file( "http://www.transmissionbt.com/donate.php" );
+    }
     else if( !strcmp( action_name, "start-torrent" ) )
     {
         GtkTreeSelection * s = tr_window_get_selection( data->wind );
