@@ -7,7 +7,7 @@
  * This exemption does not extend to derived works not owned by
  * the Transmission project.
  *
- * $Id: resume.c 8588 2009-06-02 01:48:48Z charles $
+ * $Id: resume.c 9086 2009-09-10 02:21:03Z charles $
  */
 
 #include <unistd.h> /* unlink */
@@ -15,13 +15,13 @@
 #include <string.h>
 
 #include "transmission.h"
-#include "session.h"
 #include "bencode.h"
 #include "completion.h"
 #include "fastresume.h"
 #include "peer-mgr.h" /* pex */
 #include "platform.h" /* tr_getResumeDir */
 #include "resume.h"
+#include "session.h"
 #include "torrent.h"
 #include "utils.h" /* tr_buildPath */
 
@@ -58,6 +58,11 @@
 #define KEY_PROGRESS_MTIMES   "mtimes"
 #define KEY_PROGRESS_BITFIELD "bitfield"
 
+enum
+{
+    MAX_REMEMBERED_PEERS = 200
+};
+
 static char*
 getResumeFilename( const tr_torrent * tor )
 {
@@ -73,28 +78,54 @@ getResumeFilename( const tr_torrent * tor )
 ***/
 
 static void
-savePeers( tr_benc *          dict,
-           const tr_torrent * tor )
+savePeers( tr_benc * dict, const tr_torrent * tor )
 {
-    tr_pex * pex = NULL;
-    int count = tr_peerMgrGetPeers( (tr_torrent*) tor, &pex, TR_AF_INET );
+    int count;
+    tr_pex * pex;
 
+    count = tr_peerMgrGetPeers( (tr_torrent*) tor, &pex, TR_AF_INET, MAX_REMEMBERED_PEERS );
     if( count > 0 )
         tr_bencDictAddRaw( dict, KEY_PEERS, pex, sizeof( tr_pex ) * count );
-
     tr_free( pex );
-    pex = NULL;
-    
-    count = tr_peerMgrGetPeers( (tr_torrent*) tor, &pex, TR_AF_INET6 );
+
+    count = tr_peerMgrGetPeers( (tr_torrent*) tor, &pex, TR_AF_INET6, MAX_REMEMBERED_PEERS );
     if( count > 0 )
         tr_bencDictAddRaw( dict, KEY_PEERS6, pex, sizeof( tr_pex ) * count );
-    
+
     tr_free( pex );
 }
 
+static tr_bool
+tr_isPex( const tr_pex * pex )
+{
+    return tr_isAddress( &pex->addr )
+        && ( pex->flags & 3 ) == pex->flags;
+}
+
+static int
+addPeers( tr_torrent * tor, const uint8_t * buf, int buflen )
+{
+    int i;
+    int numAdded = 0;
+    const int count = buflen / sizeof( tr_pex );
+
+    for( i=0; i<count && numAdded<MAX_REMEMBERED_PEERS; ++i )
+    {
+        tr_pex pex;
+        memcpy( &pex, buf + ( i * sizeof( tr_pex ) ), sizeof( tr_pex ) );
+        if( tr_isPex( &pex ) )
+        {
+            tr_peerMgrAddPex( tor, TR_PEER_FROM_CACHE, &pex );
+            ++numAdded;
+        }
+    }
+
+    return numAdded;
+}
+
+
 static uint64_t
-loadPeers( tr_benc *    dict,
-           tr_torrent * tor )
+loadPeers( tr_benc * dict, tr_torrent * tor )
 {
     uint64_t        ret = 0;
     const uint8_t * str;
@@ -102,29 +133,15 @@ loadPeers( tr_benc *    dict,
 
     if( tr_bencDictFindRaw( dict, KEY_PEERS, &str, &len ) )
     {
-        int       i;
-        const int count = len / sizeof( tr_pex );
-        for( i = 0; i < count; ++i )
-        {
-            tr_pex pex;
-            memcpy( &pex, str + ( i * sizeof( tr_pex ) ), sizeof( tr_pex ) );
-            tr_peerMgrAddPex( tor, TR_PEER_FROM_CACHE, &pex );
-        }
-        tr_tordbg( tor, "Loaded %d IPv4 peers from resume file", count );
+        const int numAdded = addPeers( tor, str, len );
+        tr_tordbg( tor, "Loaded %d IPv4 peers from resume file", numAdded );
         ret = TR_FR_PEERS;
     }
-    
+
     if( tr_bencDictFindRaw( dict, KEY_PEERS6, &str, &len ) )
     {
-        int       i;
-        const int count = len / sizeof( tr_pex );
-        for( i = 0; i < count; ++i )
-        {
-            tr_pex pex;
-            memcpy( &pex, str + ( i * sizeof( tr_pex ) ), sizeof( tr_pex ) );
-            tr_peerMgrAddPex( tor, TR_PEER_FROM_CACHE, &pex );
-        }
-        tr_tordbg( tor, "Loaded %d IPv6 peers from resume file", count );
+        const int numAdded = addPeers( tor, str, len );
+        tr_tordbg( tor, "Loaded %d IPv6 peers from resume file", numAdded );
         ret = TR_FR_PEERS;
     }
 
@@ -481,10 +498,12 @@ tr_torrentSaveResume( const tr_torrent * tor )
     tr_benc top;
     char *  filename;
 
-    if( !tor )
+    if( !tr_isTorrent( tor ) )
         return;
 
-    tr_bencInitDict( &top, 14 );
+    tr_tordbg( tor, "Saving .resume file for \"%s\"", tor->info.name );
+
+    tr_bencInitDict( &top, 32 ); /* arbitrary "big enough" number */
     tr_bencDictAddInt( &top, KEY_ACTIVITY_DATE,
                        tor->activityDate );
     tr_bencDictAddInt( &top, KEY_ADDED_DATE,
@@ -522,12 +541,13 @@ static uint64_t
 loadFromFile( tr_torrent * tor,
               uint64_t     fieldsToLoad )
 {
-    int64_t      i;
+    int64_t  i;
     const char * str;
-    uint64_t     fieldsLoaded = 0;
-    char *       filename;
-    tr_benc      top;
-    tr_bool      boolVal;
+    uint64_t fieldsLoaded = 0;
+    char * filename;
+    tr_benc top;
+    tr_bool boolVal;
+    const tr_bool  wasDirty = tor->isDirty;
 
     filename = getResumeFilename( tor );
 
@@ -637,10 +657,15 @@ loadFromFile( tr_torrent * tor,
 
     if( fieldsToLoad & TR_FR_SPEEDLIMIT )
         fieldsLoaded |= loadSpeedLimits( &top, tor );
-    
+
     if( fieldsToLoad & TR_FR_RATIOLIMIT )
         fieldsLoaded |= loadRatioLimits( &top, tor );
 
+    /* loading the resume file triggers of a lot of changes,
+     * but none of them needs to trigger a re-saving of the
+     * same resume information... */
+    tor->isDirty = wasDirty;
+
     tr_bencFree( &top );
     tr_free( filename );
     return fieldsLoaded;
